from pathlib import Path
import math
from copy import copy
import numpy as np

import scipy.signal as sig
from scipy.signal._upfirdn_apply import _output_len
import matplotlib.pyplot as plt

chanstocoef = {64: "kaiser64.csv"}


def pfbresponse(taps, nchans, fs):
    """Creates the frequency response of a pfb given the taps.

    Parameters
    ----------
    taps : str or ndarray
        If string will treat it as a csv file. If array treats it as taps to the filter.
    nchans : int
        The number of channesl for the pfb.
    fs : float
        Original sampling frequency of the signal in Hz.

    Returns
    -------
    freq_ar : ndarray
        Frequency vector in Hz.
    filt_dict : dict
        Keys are filter number, values are frequency response in dB
    """
    # mod_path = Path(__file__).parent.parent
    #
    # coeffile = mod_path.joinpath('coeffs',chanstocoef[nchans])
    if isinstance(taps, str):
        pfb_coefs = np.genfromtxt(str(taps), delimeter=",")
    else:
        pfb_coefs = taps
    b = pfb_coefs / np.sqrt(np.sum(np.power(pfb_coefs, 2)))

    nfreq = 2**11
    [_, h] = sig.freqz(b, 1, nfreq, fs=fs, whole=True)
    hdb = np.fft.fftshift(20 * np.log10(np.abs(h)))
    hdb = hdb - np.nanmax(hdb)
    freq_ar = np.fft.fftshift(np.fft.fftfreq(nfreq, d=1.0 / fs))

    # hpow = fftshift(20*np.log10(np.abs(h)));
    nsamps = nfreq // nchans
    filt_dict = {0: hdb}
    nplot = min(5, nchans // 2)

    for i in range(-nplot, nplot):
        filt_dict[i] = np.roll(hdb, i * nsamps)
    return freq_ar, filt_dict


def prefix(num):
    """Given a number will give you the metric prefix, symbol and muliplier, e.g. kilo, Mega etc.

    Parameters
    ----------
    num : float
        Number to be analyzed.

    Returns
    -------
    list :
        A list of the multiplier, symbol and prefix strings.
    """

    outex = int(np.floor(np.log10(np.abs(num)) - 0.3))

    prefix_dict = {
        0: [1.0, "", ""],
        3: [1e-3, "k", "kilo"],
        6: [1e-6, "M", "Mega"],
        9: [1e-9, "G", "Giga"],
    }
    return prefix_dict[outex]


def plotresponse(freq_ar, filt_dict, nchans):
    """Creates a figure and axis to plot frequency response of pfb

    First axis is a single pfb channel response, the bottom is the 5 lowest frequency.

    Parameters
    ----------
    freq_ar : ndarray
        Frequency vector in Hz.
    filt_dict : dict
        Keys are filter number, values are frequency response in dB
    nchans : int
        The number of channesl for the pfb.

    Returns
    -------
    fig : figure
        Image of frequency response.
    axs : ndarray
        List of the axes that teh frequency response plotted on.
    """
    fig, axs = plt.subplots(2, 1, figsize=(6, 8))

    mult, ack, _ = prefix(np.abs(freq_ar).max())
    fs_h = np.max(np.abs(freq_ar))
    f_wid = 1.25 * fs_h / nchans

    filtnum = 0
    hpow = filt_dict[filtnum]
    hplot = axs[0].plot(freq_ar * mult, hpow, linewidth=3)[0]
    axs[0].set_xlim([-f_wid, f_wid])
    axs[0].set_ylim([-65, 5])
    axs[0].set_xlabel("Frequency {0}Hz".format(ack))
    axs[0].set_ylabel("Magnitude dB")
    axs[0].grid(True)
    axs[0].set_title("Single Filter")

    hlist = []
    name_list = []
    for filtnum, hpow in filt_dict.items():
        hplot = axs[1].plot(freq_ar * mult, hpow, linewidth=3)[0]
        hlist.append(hplot)
        name_list.append("Channel {}".format(filtnum))
        axs[1].set_xlim([-3 * f_wid, 3 * f_wid])
        axs[1].set_ylim([-65, 5])
        axs[1].set_xlabel("Frequency {0}Hz".format(ack))
        axs[1].set_ylabel("Magnitude dB")
        axs[1].set_title("Multiple Filters")

    axs[1].legend(hlist, name_list)
    axs[1].grid(True)

    return fig, axs


def pfb_dec_simp(data, nchans, coeffs):
    """
    Parameters
    ----------
    data : ndarray
        A numpy array to be processed.
    nchans : int
        Number of output channes
    coefs : ndarray
        Filter coefficients in nchan x tpc array.

    Returns
    -------
    xout : ndarray
        The output from the polyphase.
    """
    remove_sub = False
    if data.ndim == 1:
        data = data[:, np.newaxis]
        remove_sub = True

    n_samps, subchan = data.shape
    nout = n_samps // nchans + (n_samps % nchans > 0)

    h = coeffs.copy()
    nchans_check, M_c = h.shape
    if nchans_check != nchans:
        raise ValueError(
            "Filter coefficients axis 0 size must equal number of pfb channels."
        )

    n_pre_remove = (M_c) // 2
    W = int(math.ceil(n_samps / M_c / nchans))
    # Array to be filled up
    x_summed = np.zeros((nchans, M_c * (W + 1) - 1, subchan), dtype=np.complex64)
    nfull = nchans * W * M_c
    zfill = np.zeros(nfull - n_samps, dtype=data.dtype)
    # HACK see if I can do with without
    for isub in range(subchan):
        x_p = data[:, isub]
        x_p = np.append(x_p, zfill, axis=0)
        # make x_p frequency x time orientation
        x_p = x_p.reshape((W * M_c, nchans)).T[::-1]
        for p_i, (x_i, h_i) in enumerate(zip(x_p, h)):
            # Use correlate for filtering. Due to orientation of how filter is broken up.
            # Also using full signal to make sure padding and removal is done right.
            x_summed[p_i, :, isub] = sig.convolve(x_i, h_i, mode="full")
    # now fchan x time x phychan
    xout = np.fft.ifft(x_summed, n=nchans, axis=0)[
        :, n_pre_remove : (nout + n_pre_remove)
    ]
    if remove_sub:
        xout = xout[..., 0]

    return xout


def pfb_rec_simp(data, nchans, coeffs, mask, fillmethod, fillparams=[], realout=True):
    """
    Parameters
    ----------
    data : ndarray
        A numpy array to be processed.
    nchans : int
        Number of output channes
    coefs : ndarray
        Filter coefficients in nchan x tpc array.

    Returns
    -------
    rec_array : ndarray
        The output from the polyphase synthesis.
    """

    remove_sub = False
    if data.ndim == 2:
        data = data[..., np.newaxis]
        remove_sub = True

    _, ntime, subchan = data.shape
    shp = (nchans, ntime, subchan)
    if fillmethod == "noise":
        if fillparams:
            npw = fillparams[0]
        else:
            npw = np.nanmedian(data.flatten().real ** 2 + data.flatten().imag ** 2)
            npw = npw / np.log(2)
        d1r = np.random.randn(*shp, dtype=data.dtype)
        d1i = np.random.randn(*shp, dtype=data.dtype)
        d1 = d1r + 1j * d1i

        d1 = np.sqrt(npw / 2) * d1
        rec_input = d1
    elif fillmethod == "value":
        if fillparams:
            val = fillparams[0]
        else:
            val = 0.0
        rec_input = val * np.ones(shp, dtype=data.dtype)
    else:
        rec_input = np.zeros(shp, dtype=data.dtype)

    rec_input[mask] = data

    out_data = np.fft.ifft(rec_input, n=nchans, axis=0)
    if realout:
        out_data = out_data.real
    n_samps = ntime * nchans

    rec_array = np.zeros((n_samps, subchan), dtype=out_data.dtype)

    h = coeffs.copy()
    nchans_check, M_c = h.shape
    if nchans_check != nchans:
        raise ValueError(
            "Filter coefficients axis 0 size must equal number of pfb channels."
        )

    n_pre_remove = (M_c - 1) // 2
    # Number of data samples per channel
    W = int(math.ceil(n_samps / M_c / nchans))
    # Array to be filled up
    x_summed = np.zeros((nchans, M_c * (W + 1) - 1, subchan), dtype=rec_array.dtype)
    nfull = nchans * W * M_c

    zfill = np.zeros((nchans, (nfull - n_samps) // nchans), dtype=rec_array.dtype)

    # import ipdb
    # ipdb.set_trace()
    for isub in range(subchan):
        x_p = out_data[:, :, isub]
        # x_odd = np.fft.fftshift(x_p[:,1::2],axes=0)
        # x_p[:,1::2] = x_odd
        x_p = np.append(x_p, zfill, axis=1)
        for p_i, (x_i, h_i) in enumerate(zip(x_p, h)):
            # Use correlate for filtering. Due to orientation of how filter is broken up.
            # Also using full signal to make sure padding and removal is done right.
            x_summed[p_i, :, isub] = sig.convolve(x_i, h_i, mode="full")

    for isub in range(subchan):
        # x_out = np.fft.fftshift(np.fft.fft(x_summed[:, n_pre_remove:(n_samps//nchans)+n_pre_remove, isub].T,axis=1).real,axes=1)
        # rec_array[:, isub] = x_out.flatten()
        x_out = x_summed[:, n_pre_remove : ntime + n_pre_remove, isub].T
        rec_array[:, isub] = x_out.flatten()

    if remove_sub:
        rec_array = rec_array[..., 0]

    return rec_array


def pfb_decompose(data, nchans, coefs, mask):
    """Polyphase filter function

    Takes the sampled and applies polyphase filter bank to channelize frequency content. Padding for filter is similar to scipy.signal.resample_poly, so output samples are shifted toward middle of the array. The only channels kept are those list in the mask variable.

    Parameters
    ----------
    data : ndarray
        A numpy array to be processed.
    nchans : int
        Number of output channes
    coefs : ndarray
        Filter coefficients
    mask : ndarray
        List of channels to be kept

    Returns
    -------
    xout : ndarray
        The output from the polyphase.
    """

    if data.ndim == 1:
        data = data[:, np.newaxis]
    n_samps, subchan = data.shape
    M = coefs.shape[0]
    # Determine padding for filter
    nout = n_samps // nchans + (n_samps % nchans > 0)
    # Figure out half length
    if M % 2:
        h_len = (M - 1) // 2
    else:
        h_len = M // 2

    n_pre_pad = nchans - (h_len % nchans)
    n_post_pad = 0

    n_pre_remove = (h_len + n_pre_pad) // nchans
    # Make sure you have enough samples.
    while (
        _output_len(len(coefs) + n_pre_pad + n_post_pad, n_samps, 1, nchans)
        < nout + n_pre_remove
    ):
        n_post_pad += 1
    # Make sure length of filter will be multiple of nchans
    n_post_pad += nchans - ((M + n_pre_pad + n_post_pad) % nchans)
    h_dt = coefs.dtype
    h = np.concatenate(
        [np.zeros(n_pre_pad, dtype=h_dt), coefs, np.zeros(n_post_pad, dtype=h_dt)]
    )
    n_start = (M - 1 + n_pre_pad) // 2 // nchans + (
        (((M - 1 + n_pre_pad) // 2) % nchans) > 0
    )
    # Number of filter coefficients per channel
    M_c = (M + n_pre_pad + n_post_pad) // nchans
    # Reshape filter
    h = h.reshape((nchans, M_c), order="F")[:, ::-1]
    # Number of data samples per channel
    W = int(math.ceil(n_samps / M_c / nchans))
    # Array to be filled up
    x_summed = np.zeros((nchans, M_c * W, subchan), dtype=data.dtype)
    # x_summed = np.zeros((nchans, M_c * (W + 1) - 1, subchan), dtype=np.complex64)
    nfull = nchans * W * M_c
    zfill = np.zeros(nfull - n_samps, dtype=data.dtype)
    # HACK see if I can do with without
    for isub in range(subchan):
        x_p = data[:, isub]
        x_p = np.append(x_p, zfill, axis=0)
        # make x_p frequency x time orientation
        x_p = x_p.reshape((nchans, W * M_c), order="F")
        for p_i, (x_i, h_i) in enumerate(zip(x_p, h)):
            # Use correlate for filtering. Due to orientation of how filter is broken up.
            # Also using full signal to make sure padding and removal is done right.
            x_summed[p_i, :, isub] = sig.lfilter(
                h_i,
                1,
                x_i,
            )
    # now fchan x time x phychan
    xout = np.fft.ifft(x_summed, n=nchans, axis=0)[
        mask, :  # n_pre_remove : (nout + n_pre_remove)
    ]
    return xout


def pfb_reconstruct(data, nchans, coefs, mask, fillmethod, fillparams=[], realout=True):
    """Simple PFB reconstruction

    Parameters
    ----------
    data : ndarray
        A numpy array to be processed.
    nchans : int
        Number of output channes
    coefs : ndarray
        Filter coefficients
    mask : ndarray
        List of channels to be kept
    fillmethod : string
        Type of filled in the data before the reconstrution.
    fillparams : list
        Parameters for filling in empty data

    Returns
    -------
    rec_array : ndarray
        The output from the polyphase synthesis.
    """

    if data.ndim == 2:
        data = data[:, :, np.newaxis]
    _, ntime, subchan = data.shape

    shp = (nchans, ntime, subchan)
    if fillmethod == "noise":
        if fillparams:
            npw = fillparams[0]
        else:
            npw = np.nanmedian(data.flatten().real ** 2 + data.flatten().imag ** 2)
            npw = npw / np.log(2)
        d1 = np.random.randn(shp, dtype=data.dtype) + 1j * np.random.randn(
            shp, dtype=data.dtype
        )
        d1 = np.sqrt(npw / 2) * d1
        rec_input = d1
    elif fillmethod == "value":
        if fillparams:
            val = fillparams[0]
        else:
            val = 0.0
        rec_input = val * np.ones(shp, dtype=data.dtype)
    else:
        rec_input = np.zeros(shp, dtype=data.dtype)

    rec_input[mask] = data

    out_data = np.fft.fft(rec_input, n=nchans, axis=0)

    rec_array = np.zeros((ntime * nchans, subchan), dtype=out_data.dtype)

    M = coefs.shape[0]
    # Determine padding for filter

    # Figure out half length
    if M % 2:
        h_len = (M - 1) // 2
    else:
        h_len = M // 2

    n_pre_pad = nchans - (h_len % nchans)
    n_pre_pad = 0
    n_post_pad = 0

    n_pre_remove = (h_len + n_pre_pad) // nchans
    n_samps = ntime * nchans

    # Make sure you have enough samples.
    while (
        _output_len(len(coefs) + n_pre_pad + n_post_pad, n_samps, 1, nchans)
        < ntime + n_pre_remove
    ):
        n_post_pad += 1

    # Make sure length of filter will be multiple of nchans
    n_post_pad += nchans - ((M + n_pre_pad + n_post_pad) % nchans)
    n_post_pad = n_post_pad % nchans
    h_dt = coefs.dtype

    h_full = np.concatenate(
        [
            np.zeros(n_pre_pad, dtype=data.dtype),
            coefs.astype(data.dtype),
            np.zeros(n_post_pad, dtype=data.dtype),
        ]
    )
    # Number of filter coefficients per channel
    M_c = (M + n_pre_pad + n_post_pad) // nchans
    # Reshape filter
    h_full = h_full.reshape((nchans, M_c), order="F")

    # Number of data samples per channel
    W = int(math.ceil(n_samps / M_c / nchans))
    # Array to be filled up
    x_summed = np.zeros((nchans, M_c * W, subchan), dtype=rec_array.dtype)
    nfull = nchans * W * M_c

    nadd = (nfull - n_samps) // nchans
    zfill = np.zeros((nchans, nadd), dtype=rec_array.dtype)

    # import ipdb
    # ipdb.set_trace()
    for isub in range(subchan):
        x_p = out_data[:, :, isub]
        # x_odd = np.fft.fftshift(x_p[:,1::2],axes=0)
        # x_p[:,1::2] = x_odd
        x_p = np.append(x_p, zfill, axis=1)
        for p_i, (x_i, h_i) in enumerate(zip(x_p, h_full)):
            # Use correlate for filtering. Due to orientation of how filter is broken up.
            # Also using full signal to make sure padding and removal is done right.
            x_summed[p_i, :, isub] = sig.lfilter(h_i, 1, x_i)

    for isub in range(subchan):
        # x_out = np.fft.fftshift(np.fft.fft(x_summed[:, n_pre_remove:(n_samps//nchans)+n_pre_remove, isub].T,axis=1).real,axes=1)
        # rec_array[:, isub] = x_out.flatten()
        x_out = x_summed[
            :, :, isub
        ]  # n_pre_remove: (n_samps // nchans) + n_pre_remove, isub]
        rec_array[:, isub] = x_out.flatten(order="F")[: (nfull - nadd * nchans)]

    if realout:
        rec_array = rec_array.real
    return rec_array


# %%
def npr_analysis(xin, nchans, coeffs):
    """PFB analysis using near perfect reconcsturction method.

    Performs a poly phase filter bank channelization that allows for near perfect reconstruction. This is done by creating overlaping channels to avoid any lossed from the filtering. The method is based off of this: Wessel Lubberhuizen (2024). Near Perfect Reconstruction Polyphase Filterbank (https://www.mathworks.com/matlabcentral/fileexchange/15813-near-perfect-reconstruction-polyphase-filterbank), MATLAB Central File Exchange

    Parameters
    ----------
    xin : ndarray
        The data that will be channelized along its first dimention.
    nchans : int
        Number of channels for the polyphase.
    coeffs : ndarray
        Filter coefficients, first dimension must be nchans//2.

    Returns
    -------
    yout : ndarray
        The output from the polyphase channelizer in the shape of nchans by xin.shape[0]/(nchans/2)
    """

    nfft = nchans // 2
    remove_sub = False
    if xin.ndim == 1:
        xin = xin[:, np.newaxis]
        remove_sub = True

    n_samps, subchan = xin.shape
    nout = n_samps // nfft + (n_samps % nfft > 0)

    h = coeffs.copy()

    nchans_check, M_c = h.shape
    if nchans_check != nfft:
        raise ValueError(
            "Filter coefficients axis 0 size must equal number of pfb channels."
        )

    # Change xin to work with the new type
    xin = xin.astype(np.complex128)
    # since xin has a sub channel use this
    # xin_osc = osc[:,np.newaxis]*xin

    h_comp = h.astype(xin.dtype)
    # Flip the filter coefficents in time around for the analysis
    h_comp = h_comp[:, ::-1]
    # Number of data samples per channel
    W = int(math.ceil(n_samps / M_c / nfft))
    x_summed0 = np.zeros((nfft, M_c * W, subchan), dtype=xin.dtype)
    x_summed1 = np.zeros((nfft, M_c * W, subchan), dtype=xin.dtype)

    nfull = nfft * W * M_c
    nslice = int(math.ceil(nfull / nfft))
    zfill = np.zeros(nfull - n_samps, dtype=xin.dtype)
    phi = np.mod(np.arange(nfull), 2 * nfft).astype(float) * np.pi / nfft
    osc = np.exp(1j * phi).astype(np.complex128)
    # HACK see if I can do with without the loop
    for isub in range(subchan):
        x_p0 = np.append(xin[:, isub], zfill, axis=0)
        x_p1 = x_p0 * osc
        # make x_p frequency x time orientation
        x_p0 = x_p0.reshape((nfft, nslice), order="F")
        x_p1 = x_p1.reshape((nfft, nslice), order="F")
        for p_i, (x_i0, x_i1, h_i) in enumerate(zip(x_p0, x_p1, h_comp)):
            # Use lfilter
            x_summed0[p_i, :, isub] = sig.lfilter(h_i, 1, x_i0)
            x_summed1[p_i, :, isub] = sig.lfilter(
                h_i, 1, x_i1
            )  # now fchan x time x phychan
    xout0 = np.fft.ifft(x_summed0, n=nfft, axis=0) * nfft
    xout1 = np.fft.ifft(x_summed1, n=nfft, axis=0) * nfft
    xout = np.zeros((nchans, nslice, subchan), dtype=xout0.dtype)
    xout[0:nchans:2] = xout0
    xout[1:nchans:2] = xout1
    if remove_sub:
        yout = xout[..., 0]

    return yout


# %%
def npr_synthesis(yin, nchans, coeffs, realout=False):
    """PFB synthesis using near perfect reconcsturction method.

    Performs a poly phase filter bank synthesis that allows for near perfect reconstruction. The method is based off of this: Wessel Lubberhuizen (2024). Near Perfect Reconstruction Polyphase Filterbank (https://www.mathworks.com/matlabcentral/fileexchange/15813-near-perfect-reconstruction-polyphase-filterbank), MATLAB Central File Exchange

    Parameters
    ----------
    yin : ndarray
        The data that will be channelized along its first dimension.
    nchans : int
        Number of channels for the polyphase.
    coeffs : ndarray
        Filter coefficients, first dimension must be nchans//2.
    realout : bool
        Will output a real signal if set to true.

    Returns
    -------
    xout : ndarray
        The output from the polyphase synthesis.
    """
    nfft = nchans // 2
    remove_sub = False
    if yin.ndim == 2:
        yin = yin[..., np.newaxis]
        remove_sub = True

    _, ntime, subchan = yin.shape

    shp = (nchans, ntime, subchan)
    yin_e = yin[0:nchans:2]
    yin_o = yin[1:nchans:2]

    xe = np.fft.fft(yin_e, n=nfft, axis=0) * nfft
    xo = np.fft.fft(yin_o, n=nfft, axis=0) * nfft

    n_samps = ntime * nfft

    h = coeffs.copy()
    nchans_check, M_c = h.shape
    if nchans_check != nfft:
        raise ValueError(
            "Filter coefficients axis 0 size must equal number of pfb channels."
        )

    # Number of data samples per channel
    W = int(math.ceil(n_samps / M_c / nfft))
    # Array to be filled up
    x_summede = np.zeros((nfft, M_c * W, subchan), dtype=xe.dtype)

    x_summedo = np.zeros((nfft, M_c * W, subchan), dtype=xe.dtype)

    nfull = nfft * W * M_c
    xout = np.zeros((nfull, subchan), dtype=xe.dtype)

    zfill = np.zeros((nfft, (nfull - n_samps) // nfft), dtype=xout.dtype)
    phi = np.mod(np.arange(nfull), nchans).astype(float) * np.pi / nfft
    osterm = np.exp(-1j * phi)
    # import ipdb
    # ipdb.set_trace()
    for isub in range(subchan):
        x_pe = xe[:, :, isub]
        x_po = xo[:, :, isub]
        # x_odd = np.fft.fftshift(x_p[:,1::2],axes=0)
        # x_p[:,1::2] = x_odd
        x_pe = np.append(x_pe, zfill, axis=1)
        x_po = np.append(x_po, zfill, axis=1)
        for p_i, (x_ie, x_io, h_i) in enumerate(zip(x_pe, x_po, h)):
            # Use correlate for filtering. Due to orientation of how filter is broken up.
            # Also using full signal to make sure padding and removal is done right.
            x_summedo[p_i, :, isub] = sig.lfilter(h_i, 1, x_io)
            x_summede[p_i, :, isub] = sig.lfilter(h_i, 1, x_ie)

    for isub in range(subchan):
        # x_out = np.fft.fftshift(np.fft.fft(x_summed[:, n_pre_remove:(n_samps//nchans)+n_pre_remove, isub].T,axis=1).real,axes=1)
        # xout[:, isub] = x_out.flatten()
        x_oute = x_summede[:, : nfull // nfft, isub].T
        x_outo = x_summedo[:, : nfull // nfft, isub].T
        xout[:, isub] = x_oute.flatten() - x_outo.flatten() * osterm

    if remove_sub:
        xout = xout[..., 0]
    if realout:
        xout = xout.real
    return xout



from pathlib import Path
import scipy.signal as sig
import numpy as np
import scipy.special as scisp


def kaiser_coeffs(nchans, beta=1.7 * np.pi, pow2=True):
    """Creates a Type 1 Kaiser window with a flat passband. Adds zeros to the begining of the filter to give full samples as the delay.

    Parameters
    ----------
    nchans : int
        Number of frequency channels that the coefficients will be used for.
    beta : float
        Shape parameter for the filter.
    pow2 : bool
        Pad coefficients with zeros to the next power of 2.

    Returns
    -------
    taps : array_like
        Resulting filter taps.
    """

    ntaps = 64 * nchans
    # If you double this you can do this all with integers
    fs = nchans * 2
    if pow2:
        ntaps = int(np.power(2, np.ceil(np.log2(ntaps))))
    # Make odd number so you have type 1 filter.
    furry = sig.firwin(ntaps - 1, 1, window=("kaiser", beta), scale=True, fs=fs)
    taps = np.concatenate(([0], furry))
    return taps

    # sig.firwin(24*N, 0.5, window=('kaiser', 3*np.pi), scale=True, fs=N)


def kaiser_syn_coeffs(nchans, beta=1.7 * np.pi, pow2=True):
    """Creates a Kaiser window with a flat passband

    Parameters
    ----------
    nchans : int
        Number of frequency channels that the coefficients will be used for.
    beta : float
        Shape parameter for the filter.
    pow2 : bool
        Pad coefficients with zeros to the next power of 2.

    Returns
    -------
    taps : array_like
        Resulting filter taps.
    """

    ntaps = 64 * nchans
    # If you double this you can do this all with integers
    fs = nchans * 2
    if pow2:
        ntaps = int(np.power(2, np.ceil(np.log2(ntaps))))
    # Make odd number so you have type 1 filter.
    furry = sig.firwin(ntaps - 1, 1, window=("kaiser", beta), scale=True, fs=fs)
    taps = np.concatenate(([0], furry))
    taps = taps * nchans**2
    return taps


def kaiser_pfb_coefs(nchans, tpc=32, npr=False, beta=1.7 * np.pi):
    """Creates a Kaiser window with a flat passband with specific control for the

    Parameters
    ----------
    nchans : int
        Number of frequency channels that the coefficients will be used for.
    beta : float
        Shape parameter for the filter.
    npr : bool
        If yes then the number of rows is divided by 2 for the npr style filter.
    beta : float
        A shaping parameter for the kaiser functions. Default : 1.7*pi

    Returns
    -------
    taps : array_like
        Resulting filter taps as nchans x tpc array.
    """

    if npr:
        ncols = nchans // 2
    else:
        ncols = nchans

    ntaps = tpc * ncols
    # If you double this you can do this all with integers
    fs = nchans * 2

    # Make even number so you have type 2 filter.
    taps = sig.firwin(ntaps, 1, window=("kaiser", beta), scale=True, fs=fs)
    coeffs = taps.reshape((ncols, tpc), order="F")
    return coeffs


def createcoeffs(savedir):
    """Create a set of files for taps.

    Parameters
    ----------
    savedir : str
        Directory where the data will be saved.

    """
    chanarr = 2 ** np.arange(1, 11)
    maxchans = chanarr.max()
    maxchar = int(np.ceil(np.log10(maxchans)))
    suf_str = "{:0" + str(maxchar) + "}"
    savepath = Path(savedir)
    fstema = "kaiseranalysis" + suf_str + ".csv"
    fstems = "kaisersynthesis" + suf_str + ".csv"
    for ichans in chanarr:
        taps = kaiser_coeffs(ichans, pow2=False)

        fname = savepath.joinpath(fstema.format(ichans))
        np.savetxt(fname, taps, delimiter=",")

        if ichans >= 4:
            taps = kaiser_syn_coeffs(ichans, pow2=False) * ichans / 2
            fname = savepath.joinpath(fstems.format(ichans))
            np.savetxt(fname, taps, delimiter=",")


def rref_coef(N, L, K=None):
    """Creates a filter with the frequency response of the a root error function and sets it up for a npr filter bank. Based off the following: Wessel Lubberhuizen (2024). Near Perfect Reconstruction Polyphase Filterbank (https://www.mathworks.com/matlabcentral/fileexchange/15813-near-perfect-reconstruction-polyphase-filterbank), MATLAB Central File Exchange

    Parameters
    ----------
    N : int
        Number channels
    L : int
        Number of taps per channel.
    K : float
        A shaping factor but has default values

    Returns
    -------
    coeffs : ndarray
        A N//2 by L array of the coeffiecents.
    """
    k_dict = {
        8: 4.853,
        10: 4.775,
        12: 5.257,
        14: 5.736,
        16: 5.856,
        18: 7.037,
        20: 6.499,
        22: 6.483,
        24: 7.410,
        26: 7.022,
        28: 7.097,
        30: 7.755,
        32: 7.452,
        48: 8.522,
        64: 9.457,
        96: 10.785,
        128: 11.5,
        192: 11.5,
        256: 11.5,
    }

    if K is None:
        K = k_dict.get(L, 8.0)
    M = N // 2

    f = np.arange(0, L * M, dtype=np.float64) / (L * M)
    x = K * (2 * M * f - 0.5)
    A = np.sqrt(0.5 * scisp.erfc(x))
    a_n = len(A)
    n = np.arange(1, a_n / 2 + 1, dtype=int)
    A[-n] = np.conj(A[n])
    A[int(a_n / 2)] = 0
    B = np.fft.ifft(A)
    B = np.fft.fftshift(B).real

    return B.reshape((M, L), order="F")



#!/usr/bin/env python
"""
detectors.py
Code for constant false alarm rate and other detectors.
"""
import numpy as np
import scipy.special as sc
from pickle import FLOAT


def cfar(data, cfar_width, thresh):
    """Run the cfar detector on a data set with the same threshold unit.

    Runs a constant false alarm detector in one dimension.

    Parameters
    ----------
    data : ndarray
        Data array that will be examined.
    cfar_width : int
        Width of the CFAR array.
    thresh : float
        Threshold for making a detection.

    Returns
    -------
    outlist : list
        List of dictionaries with detections.
    """
    outlist = []
    ncells, nframes = data.shape
    # Get the median for the whole data set
    med_all = np.nanmedian(data.flatten())
    outdict = dict(cell=[], frame=[], snr=[], npow=[])
    for iframe in range(nframes):
        curframe = data[:, iframe]
        for icell in range(ncells):
            lbnd = min(ncells - cfar_width, max(icell - cfar_width // 2, 0))
            ubnd = max(cfar_width, min(icell + cfar_width // 2, ncells))
            val = curframe[icell]
            floor = max(np.nanmedian(curframe[lbnd:ubnd]), med_all)
            if val >= thresh + floor:

                # d1 = dict(cell=icell,frame=iframe,snr=val-floor,npow=floor)
                outdict["cell"].append(icell)
                outdict["frame"].append(iframe)
                outdict["snr"].append(val - floor)
                outdict["npow"].append(floor)
                # outlist.append(d1)
    return outdict


def erlang_ratio(pfa, k):
    """Get the multiplier for a detector for an Erlang distributed variable.

    Create the threshold for a erlang random variable by using inverse lower incomplete gamma functions.

    Parameters
    ----------
    pfa : float
        Probability of false alarm.
    k : int
        Shape parameter or number of exponentially distributed terms added together.

    Returns
    -------
    erlang_ratio
        Multiplier to go from median to mean.
    """

    gm_num = sc.gammaincinv(k, 1.0 - pfa)
    gm_den = sc.gammaincinv(k, 0.5)
    erlang_ratio = gm_num / gm_den

    return erlang_ratio



"""
Various types of Time-Frequency estimators.
"""

from pathlib import Path
import argparse
import scipy.io.wavfile as wavio
import matplotlib.pyplot as plt
import scipy.signal as sig
import numpy as np


def create_sti(filename, nfft, decimation, secoffset):
    """This does the calculation for the STI plot

    Creates the spectral time intensity plot of the whole file. Outputs the sti in dB/Hz. Array is frequency x time.

    Parameters
    ----------
    filename : str
        Name of the wav file.
    nfft : int
        Number of fft bins after decimation.
    decimation : int
        Decimation factor.
    secoffset : float
        Number of seconds to offset the time vector.
    Returns
    -------
    freq : array_like
        Vector of frequency values in Hz
    t_ar : array_like
        Time array in seconds.
    len_s : float
        Number of seconds of data in file.
    Sxxlist : list
        List of the spectrogram arrays in dB/Hz.
    """
    rate, data = wavio.read(filename)
    len_s = float(data.shape[0]) / rate
    freq = np.fft.rfftfreq(nfft, 1.0 / rate)
    if data.ndim == 1:
        if decimation != 1:
            data = sig.decimate(data, decimation)
            rate = rate / decimation

        ntime = data.shape[0] // nfft
        t_ar = (np.arange(ntime) + 0.5) * nfft / rate
        datamat = data[: ntime * nfft].reshape(ntime, nfft)
        Sxx0 = np.abs(np.fft.rfft(datamat, axis=1)) ** 2 / rate
        # freq, t_ar, Sxx0 = sig.spectrogram(data, rate, nfft=nfft, scaling="density")
        Sxx0db = 10 * np.log10(Sxx0.transpose() + 1e-6)
        Sxxlist = [Sxx0db]
    elif data.ndim == 2:
        nchan = data.shape[1]
        Sxxlist = []
        if decimation != 1:
            data = sig.decimate(data, decimation, axis=0)
            rate = rate / decimation
        for ichan in range(nchan):
            ntime = data.shape[0] // nfft
            t_ar = (np.arange(ntime) + 0.5) * nfft / rate
            datamat = data[: ntime * nfft, ichan].reshape(ntime, nfft)
            Sxx0 = np.abs(np.fft.rfft(datamat, axis=1)) ** 2 / rate
            # freq, t_ar, Sxx0 = sig.spectrogram(
            #     data[:, ichan], rate, nfft=nfft, scaling="density"
            # )
            Sxx0db = 10 * np.log10(Sxx0.transpose() + 1e-12)
            Sxxlist.append(Sxx0db)
    t_ar = t_ar + secoffset
    return freq, t_ar, len_s, Sxxlist


# %% Lag Functions


def lag_product(x_in, nlag, nmean=10, numtype=np.complex64, lagtype="centered"):
    """This function will create a  lag product. It will form each lag along the first dimension and then perform sub integrations (arithmatic mean) and then apply a median across the sub integrations. This will deal with n-dimensional arrays where the first two dimensions are the process overwhich the acf is formed and the second is a set of realizations overwhich it is estimated.

    Parameters
    ----------
    x_in : ndarray
        This is a numpy array where the first two dimensions are NsxNp, where Ns is number of samples that process is varying (i.e. range gates in ISR) and Np is number of process relizations (e.g. pulses in ISR).
    nmean : int
        Number of pulses to be averaged first before median applied.
    nlag : int
        Length of the lag product formation.
    numtype : type
        numerical representaiton of the array.

    lagtype : str
        Can be centered forward or backward.

    Returns
    -------
    acf_cent : ndarray
        This is a NrxNl complex numpy array where Nr is number of range gate and Nl is number of lags.
    rg_keep : ndarray
        Indicies of the samples/range gates that will be kept after the lag formation.
    """
    # The first dimension is the where estimation is formed, the second dimension is integrated over. The other dimensions will left along and just carried through.
    (Nr, Np) = x_in.shape[:2]
    otherdims = x_in.shape[2:]
    med_int = np.arange(0, Np, nmean)
    n_start = med_int[:-1]
    n_end = med_int[1:]
    n_sub = len(n_start)
    # Make masks for each piece of data
    if lagtype == "forward":
        arback = np.zeros(nlag, dtype=int)
        arfor = np.arange(nlag, dtype=int)

    elif lagtype == "backward":
        arback = -np.arange(nlag, dtype=int)
        arfor = np.zeros(nlag, dtype=int)
    else:
        # arex = np.arange(0,N/2.0,0.5);
        arback = -np.floor(np.arange(0, nlag / 2.0, 0.5)).astype(int)
        arfor = np.ceil(np.arange(0, nlag / 2.0, 0.5)).astype(int)

    # figure out how much range space will be kept
    ap = -1 * np.min(arback)
    ep = Nr - np.nanmax(arfor)
    rg_keep = np.arange(ap, ep)
    #    wearr = (1./(N-np.tile((arfor-arback)[:,np.newaxis],(1,Np)))).astype(numtype)
    # acf_cent = np.zeros((ep-ap,N))*(1+1j)
    acf_cent = np.zeros((ep - ap, nlag, *otherdims), dtype=numtype)
    for irng, curange in enumerate(rg_keep):
        rng_ar1 = int(curange) + arback
        rng_ar2 = int(curange) + arfor
        # get all of the acfs across pulses # sum along the pulses
        acf_tmp = np.conj(x_in[rng_ar1, :]) * x_in[rng_ar2, :]  # *wearr
        acf_sub = np.zeros((n_sub, nlag, *otherdims), dtype=numtype)
        for i_ind, (ibeg, iend) in enumerate(zip(n_start, n_end)):
            tmp_acf = acf_tmp[:, slice(ibeg, iend)]
            acf_sub[i_ind] = np.nanmean(tmp_acf, axis=1)
        acf_ave = np.nanmedian(acf_sub, axis=0)

        acf_cent[irng, :] = acf_ave  # might need to transpose this
    return acf_cent, rg_keep


def make_sum_rule(nlag, lagtype="centered", srule=None):
    """This function will return the sum rule.

    Parameter
    ---------
    nlag : int
        Number of lags in the ACF.
    lagtype : str
        Type of lags, centered, forward or backward.

    Returns
    -------
    sumrule : ndarray
        A 2 x nlag numpy array that holds the summation rule.
    """
    if lagtype == "forward":
        arback = -np.arange(nlag, dtype=int)
        arforward = np.zeros(nlag, dtype=int)
    elif lagtype == "backward":
        arback = np.zeros(nlag, dtype=int)
        arforward = np.arange(nlag, dtype=int)
    elif lagtype == "centered":
        arback = -np.ceil(np.arange(0, nlag / 2.0, 0.5)).astype(int)
        arforward = np.floor(np.arange(0, nlag / 2.0, 0.5)).astype(int)
    elif lagtype == "barker":
        arback = np.zeros(1, dtype=int)
        arforward = np.zeros(1, dtype=int)
    sumrule = np.array([arback, arforward])
    return sumrule


def make_acf(
    x_in, nlag, nmean=10, numtype=np.complex64, lagtype="centered", srule="centered"
):
    """Performs the full ACF estimation including application of the summation rule to equalize the lag statistics. This will deal with n-dimensional arrays where the first two dimensions are the process overwhich the acf is formed and the second is a set of realizations overwhich it is estimated.

    Parameters
    ----------
    x_in : ndarray
        This is a numpy array where the first two dimensions are NsxNp, where Ns is number of samples that process is varying (i.e. range gates in ISR) and Np is number of process relizations (e.g. pulses in ISR).
    nmean : int
        Number of pulses to be averaged first before median applied.
    nlag : int
        Length of the lag product formation.
    numtype : type
        numerical representaiton of the array.
    lagtype : str
        Can be centered forward or backward.
    srule : str

    Returns
    -------
    acf_est : ndarray
        This is a NrxNl complex numpy array where Nr is number of range gate and Nl is number of lags.
    rg_keep : ndarray
        Indicies of the samples/range gates that will be kept after the lag formation.
    """
    assert nlag >= 1, "nlag must be greater or equal to than one."

    sumrule = make_sum_rule(nlag, lagtype, srule)
    y_out, rng_k1 = lag_product(x_in, nlag, nmean, numtype, lagtype)
    n_rg = y_out.shape[0]
    otherdims = x_in.shape[2:]
    minrg = -1 * sumrule[0].min()
    maxrg = n_rg - sumrule[1].max()
    n_rg_out = maxrg - minrg
    rng_k2 = rng_k1[minrg:maxrg]
    acf_est = np.zeros((n_rg_out, nlag, *otherdims), dtype=numtype)
    for inum, irg in enumerate(range(minrg, maxrg)):
        for ilag in range(nlag):
            cur_sr = sumrule[:, ilag]
            r_sl = slice(irg + cur_sr[0], irg + cur_sr[1]+1)

            # perform a sum on the lags instead of a averaging otherwise you have to weight a window on the output.
            acf_est[inum,ilag] = np.nansum(y_out[r_sl, ilag], axis=0)
    return acf_est, rng_k2


def sub_int(data, nsubs, axis=0):
    """Perform sub integration (mean) and then median on the data.

    Parameters
    ----------
    data : ndarray
        Data array to be averaged over.
    nsubs : int
        Sub integration periods
    axis : int
        Axis over which the data is averaged.

    Returns
    -------
    data_int : ndarray
        Data array after the averaging
    """
    dshape = data.shape
    n_els = dshape[axis]
    tshape = list(dshape)
    tshape[axis] = nsubs
    ar1 = np.linspace(0, n_els, nsubs, dtype=int, endpoint=False)
    ar2 = np.roll(ar1, -1)
    ar2[-1] = n_els
    int_points = np.column_stack((ar1, ar2))
    tmp_data = np.empty(tshape, dtype=data.dtype)

    for i_int, iloc in enumerate(int_points):
        cur_d = data.take(range(*iloc), axis=axis)
        tmp_data[i_int] = np.nanmean(cur_d, axis=axis)
    data_int = np.nanmedian(tmp_data, axis=axis)

    return data_int


# %% Barker Codess
def gen_bark(blen):
    """This function will output a Barker code pulse.

    Parameters
    ----------
    blen : int
        An integer for number of bauds in barker code.

    Returns
    -------
    outar : ndarray
        A blen length numpy array.
    """
    bdict = {
        1: [-1],
        2: [-1, 1],
        3: [-1, -1, 1],
        4: [-1, -1, 1, -1],
        5: [-1, -1, -1, 1, -1],
        7: [-1, -1, -1, 1, 1, -1, 1],
        11: [-1, -1, -1, 1, 1, 1, -1, 1, 1, -1, 1],
        13: [-1, -1, -1, -1, -1, 1, 1, -1, -1, 1, -1, 1, -1],
    }
    outar = np.array(bdict[blen])
    outar.astype(np.float64)
    return outar


def barker_lag(x_in, numtype=None, pulse=gen_bark(13)):
    """This will process barker code data by filtering it with a barker code pulse and then sum up the pulses.

    Paramaters
    ----------
    x_in : ndarray
        A complex numpy array size NpxNs where Np is the number of pulses and
    numtype : type
        The type used for processing the data.
    pulse : ndarray
        The barkercode pulse.
    Returns
    -------
    outdata : ndarray
        A Nrx1 size numpy array that holds the processed data. Nr is the number of range gates
    """

    if numtype is None:
        numtype = x_in.dtype

    # It will be assumed the data will be pulses vs rangne
    x_in = x_in.transpose()
    (Nr, Np) = x_in.shape
    pulsepow = np.power(np.absolute(pulse), 2.0).sum()
    # Make matched filter
    filt = np.fft.fft(pulse[::-1] / np.sqrt(pulsepow), n=Nr)
    filtmat = np.repeat(filt[:, np.newaxis], Np, axis=1)
    rawfreq = np.fft.fft(x_in, axis=0)
    outdata = np.fft.ifft(filtmat * rawfreq, axis=0)

    outdata = outdata * outdata.conj()
    outdata = np.sum(outdata, axis=-1)
    # increase the number of axes
    return outdata[len(pulse) - 1 :, np.newaxis]



from .estimators import create_sti, make_acf, make_sum_rule, barker_lag, lag_product,sub_int
from .detectors import cfar
from .filtertools import kaiser_coeffs, kaiser_syn_coeffs, kaiser_pfb_coefs, rref_coef
from .pfb import (
    pfb_decompose,
    pfb_reconstruct,
    pfbresponse,
    pfb_dec_simp,
    pfb_rec_simp,
    npr_analysis,
    npr_synthesis,
)

from . import _version

__version__ = _version.get_versions()["version"]



#!/usr/bin/env python
"""
setup.py
This is the setup file for the ARS signal processing python package

@author: John Swoboda
"""
from pathlib import Path
from setuptools import setup, find_packages
import versioneer

req = ["scipy", "numpy", "matplotlib", "digital_rf"]
scripts = ["bin/drf_decompose.py"]


config = dict(
    description="Processing and Plotting of ",
    author="John Swoboda",
    url="https://github.com/MIT-Adaptive-Radio-Science/sigprocpython",
    version='1.0.1',
    cmdclass=versioneer.get_cmdclass(),
    install_requires=req,
    python_requires=">=3.0",
    packages=find_packages(),
    scripts=scripts,
    name="mitarspysigproc",
    package_data={"mitarspysigproc": ["coeffs/*.csv"]},
)

curpath = Path(__file__)
testpath = curpath.joinpath("Testdata")
try:
    curpath.mkdir(parents=True, exist_ok=True)
    print("created {}".format(testpath))
except OSError:
    pass


setup(**config)



#!/usr/bin/env python
"""

"""
import argparse
from pathlib import Path

import sys
import shutil
import math
import numpy as np
import digital_rf as drf

from mitarspysigproc import pfb_decompose, kaiser_coeffs



def parse_command_line(str_input=None):
    """This will parse through the command line arguments

    Function to go through the command line and if given a list of strings all also output a namespace object.

    Parameters
    ----------
    str_input : list
        A list of strings from the command line. If none will read from commandline. Can just take command line inputs and do a split() on them.

    Returns
    -------
    input_args : Namespace
        An object holding the input arguments wrt the variables.
    """
    scriptname = Path(sys.argv[0]).name

    formatter = argparse.RawDescriptionHelpFormatter(scriptname)
    width = formatter._width
    title = "DRF PFB"
    shortdesc = (
        "Run a polyphase filter bank and break up data into frequency channels. "
    )
    desc = "\n".join(
        (
            "*" * width,
            "*{0:^{1}}*".format(title, width - 2),
            "*{0:^{1}}*".format("", width - 2),
            "*{0:^{1}}*".format(shortdesc, width - 2),
            "*" * width,
        )
    )
    # desc = "This is the run script for SimVSR."
    # if str_input is None:
    parser = argparse.ArgumentParser(
        description=desc, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    # else:
    #     parser = argparse.ArgumentParser(str_input)

    parser.add_argument(
        "-p",
        "--path",
        dest="path",
        required=True,
        help="Name of the digital RF data set directory.",
    )
    parser.add_argument(
        "-n",
        "--pfbchans",
        dest="pfbchans",
        help="Number of pfb channels",
        required=True,
        type=int,
    )
    parser.add_argument(
        "-c",
        "--channame",
        dest="channame",
        help="Name of channel with sub channel seperated by a colon.",
        default=None,
    )

    parser.add_argument(
        "-o",
        "--outdir",
        dest="outdir",
        required=False,
        default="",
        type=str,
        help="Altitude in km for the average orbit.",
    )
    parser.add_argument(
        "-m", "--mask", dest="mask", required=False, type=str, default=""
    )
    parser.add_argument(
        "-s",
        "--starttime",
        dest="starttime",
        default=None,
        help="Use the provided start time instead of the first time in the data. format is ISO8601: 2015-11-01T15:24:00Z",
    )
    parser.add_argument(
        "-e",
        "--endtime",
        dest="endtime",
        default=None,
        help="Use the provided end time for the plot. format is ISO8601: 2015-11-01T15:24:00Z",
    )

    if str_input is None:
        return parser.parse_args()
    return parser.parse_args(str_input)


def setuppfbout(datadir, chan, signal_params, pfb_params, st):
    """Sets up the digital_rf writer objects for each pfb channel.

    Parameters
    ----------
    datadir : str
        Location of the director where the data will be stored.
    chan : str
        Name of the channel that will be read
    signal_params : dict
        Dictionary from the get_properties function of the drf reader.
    pfb_params : dict
        Dictionary holding info related to the pfb, filter coefiecents, number of channels, etc.
    st : datetime
        Start time of the decomposition.

    Returns
    -------
    pfb_out : dict
        Dictionary hold DigitalRFWriters.
    pfb_obj : dict
        Dictionary holding metadata writters.
    """

    datapath = Path(datadir)

    sr_nfs = signal_params["sample_rate_numerator"]
    sr_dfs = signal_params["sample_rate_denominator"]
    n_subchannels = signal_params["num_subchannels"]

    common_params = {
        "subdir_cadence_secs": 3600,  # Number of seconds of data in a subdirectory
        "file_cadence_millisecs": 10000,  # Each file will have up to 400 ms of data
        "compression_level": 0,  # no compression
        "checksum": False,  # no checksum
        "is_continuous": True,
        "num_subchannels": n_subchannels,
        "marching_periods": False,  # no marching periods when writing
        "sample_rate_numerator": sr_nfs,
    }
    # Common parameters
    sub_cadence_secs = common_params[
        "subdir_cadence_secs"
    ]  # Number of seconds of data in a subdirectory
    file_cadence_secs = (
        common_params["file_cadence_millisecs"] // 1000
    )  # Each file will have up to 400 ms of data
    compression_level = 0  # no compression
    checksum = False  # no checksum

    is_continuous = True
    marching_periods = False  # no marching periods when writing
    # get channels

    # pfb

    sr_div = pfb_params["nchans"]
    freq_map = np.fft.fftfreq(sr_div, d=float(sr_dfs) / sr_nfs)
    pfb_out = {}
    pfb_obj = {}
    ndig = int(np.log10(sr_div)) + 1
    chan_list = np.arange(sr_div)
    for ichan in chan_list:
        if not ichan in pfb_params["mask"]:
            continue
        # HACK how many channels will we need?
        iname = chan + "_" + str(ichan).zfill(ndig)
        cdir = datapath.joinpath(iname)
        if cdir.is_dir():
            shutil.rmtree(str(cdir))
        cdir.mkdir(parents=True, exist_ok=True)
        start_indx_fs = drf.util.datetime_to_timestamp(st) * sr_nfs / sr_dfs
        pfb_info = common_params.copy()
        pfb_info["directory"] = str(cdir)

        pfb_info["start_global_index"] = int(start_indx_fs / sr_div)
        pfb_info["sample_rate_denominator"] = sr_div * sr_dfs
        pfb_info["uuid_str"] = "pfb"
        pfb_info["dtype"] = np.complex64  # complex64
        pfb_info["is_complex"] = True  # complex values
        pfb_out[ichan] = drf.DigitalRFWriter(**pfb_info)

        # Create meta data
        pfbmetadir = cdir.joinpath("metadata")
        if pfbmetadir.is_dir():
            shutil.rmtree(str(pfbmetadir))
        pfbmetadir.mkdir(parents=True, exist_ok=True)

        pfb_obj[ichan] = drf.DigitalMetadataWriter(
            str(pfbmetadir),
            sub_cadence_secs,
            file_cadence_secs,
            sr_nfs,
            sr_div * sr_dfs,
            "pfb",
        )
        pfbap = pfb_params.copy()
        # cur_freq = [freq_map[ichan]] * n_subchannels
        # pfbap["center_frequencies"] = np.array(cur_freq)
        pfb_obj[ichan].write(pfb_info["start_global_index"], pfbap)
    return pfb_out, pfb_obj


def runpfb(path, pfbchans, channame, outdir="", mask="", starttime="", endtime=""):
    """Read in a Digital RF data set and decompose data using a polyphase filter bank.

    Parameters
    ----------
    path : str
        Path to the input data set.
    pfbchans : int
        Number of frequency channels the data is to be broken into
    channame : str
        Name of channel with sub channel seperated by a colon.
    outdir : str
        Directory that the decompsoed data will be written to.
    mask : str
        Which frequency channels, numbered 0 to pfbchans-1, to keep.
    starttime : str
        Start time instead of the first time in the data. format is ISO8601: 2015-11-01T15:24:00Z
    endtime : str
        End time instead of the last time in the data. format is ISO8601: 2015-11-01T15:24:00Z
    """

    # Parameter fixes
    if outdir == "":
        outdir = path.copy()

    if mask == "":
        mask = "0-{0}".format(pfbchans)
        mask_arr = np.arange(pfbchans)
    elif "-" in mask:
        beg, fin = mask.split("-")
        mask_arr = np.arange(int(beg), int(fin))
    elif "," in mask:
        maskstr = mask.split(",")
        mask_arr = np.array([int(i) for i in maskstr])

    drfObj = drf.DigitalRFReader(str(path))
    chans = drfObj.get_channels()
    if channame == "":
        chan = chans[0]
        subchan = 0
    else:
        chan, subchan = channame.split(":")
        subchan = int(subchan)

    # Get the signal properties
    signal_params = drfObj.get_properties(chan)
    st_f, et_f = drfObj.get_bounds(chan)
    sr = signal_params["samples_per_second"]

    if starttime == "":
        st = drf.util.sample_to_datetime(st_f, sr)
        st_samp = st_f
    else:
        st = drf.util.parse_identifier_to_time(starttime, sr)
        st_samp = int(drf.util.datetime_to_timestamp(st) * sr)

    if endtime == "":
        et = drf.util.sample_to_datetime(et_f, sr)
        et_samp = et_f
    else:
        et = drf.util.parse_identifier_to_time(endtime, sr)
        et_samp = int(drf.util.datetime_to_timestamp(et) * sr)

    sam_per_read = int(32 * 2**20 / 4)

    read_time = np.arange(st_samp, et_samp, sam_per_read)
    pfb_params = {
        "nchans": pfbchans,
        "coefs": kaiser_coeffs(pfbchans),
        "mask": mask_arr,
    }
    pfb_coefs = pfb_params["coefs"]

    # Normalize the coefiecents to try to keep the variances
    pfb_coefs = pfb_coefs / np.sqrt(np.sum(np.power(pfb_coefs, 2)))
    n_coefs = pfb_coefs.size
    M = n_coefs + ((n_coefs + 1) % 2)
    # Determine the amount of overlap needed after the filtering to keep things continuous between reads.

    h_len = (M - 1) // 2
    ds_fac = pfb_params["nchans"]
    h_ds = h_len // ds_fac + (h_len % ds_fac > 0)
    channel_taps = math.ceil(n_coefs / ds_fac)
    nfull = ds_fac * channel_taps
    noutlist = sam_per_read // ds_fac + (sam_per_read % ds_fac > 0)
    pfb_params["n_del"] = h_ds
    pfb_params["nout"] = noutlist
    pfb_params["coefs"] = pfb_coefs

    pfb_out, pfb_obj = setuppfbout(outdir, chan, signal_params, pfb_params, st)
    print("Decomposing {0} into {1} channels".format(path, pfbchans))
    print("Keeping channels {0}".format(mask))

    for ind, ist in enumerate(read_time):
        print("Reading {0} out of {1}".format(ind + 1, len(read_time)))
        nsamps = min(sam_per_read + M, et_samp - ist)
        x = drfObj.read_vector(ist, nsamps, chan, subchan)
        x_out = pfb_decompose(
            x, pfb_params["nchans"], pfb_params["coefs"], pfb_params["mask"]
        )
        x_out = x_out[:,channel_taps:]
        for ipchan, iwriter in pfb_out.items():
            iwriter.rf_write(x_out[ipchan].astype(iwriter.dtype))
    for ipchan, iwriter in pfb_out.items():
        iwriter.close()


if __name__ == "__main__":
    args_commd = parse_command_line()
    arg_dict = {k: v for k, v in args_commd._get_kwargs() if v is not None}
    runpfb(**arg_dict)
